"""
This Python 3.11 module implements a wrapper for map projections.

- Alexander Raichev (AR), 2013-01-25: Refactored code from release 0.3.

NOTE:

All lengths are measured in meters and all angles are measured in radians
unless indicated otherwise.
By 'ellipsoid' below, I mean an oblate ellipsoid of revolution.
"""

# *****************************************************************************
#       Copyright (C) 2013 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import third-party modules.
import pyproj

# Import standard modules.
import importlib

# Import my modules.
from vgrid.utils.rhealpixdggs.utils import wrap_longitude, wrap_latitude
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID


# Homemade map projections, as opposed to those in the PROJ.4 library.
# Remove 'healpix' and 'rhealpix' to use the PROJ.4 versions instead,
# assuming you have the *correct/patched* PROJ.4 versions.
HOMEMADE_PROJECTIONS = {"healpix", "rhealpix", "isea", "csea", "qsc"}
# HOMEMADE_PROJECTIONS = {"isea", "csea", "qsc"}


class Projection(object):
    """
    Represents a map projection of a given ellipsoid.

    INSTANCE ATTRIBUTES:

    - `ellipsoid` - An ellipsoid (Ellipsoid instance) to project.
    - `proj` - The name (string) of the map projection, either a valid PROJ.4
      projection name or a valid homemade projection name.
    - `kwargs` - Keyword arguments (dictionary) needed for the projection's
      definition, but not for the definition of the ellipsoid.  For example,
      these could be {'north_square':1, 'south_square': 2} for the rhealpix
      projection.

    EXAMPLES::

        >>> from rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
        >>> f = Projection(ellipsoid=WGS84_ELLIPSOID, proj='rhealpix', north_square=1, south_square=0)
        >>> print(tuple(x.tolist() for x in my_round(f(0, 30), 15)))
        (0.0, 3740232.8933662786)
        >>> f = Projection(ellipsoid=WGS84_ELLIPSOID, proj='cea')
        >>> print(my_round(f(0, 30), 15))
        (0.0, 3171259.315518537)

    NOTES:

    When accessing a homemade map projection assume that it can be called via
    a function g(a, e), where a is the major radius of the ellipsoid to be
    projected and e is its eccentricity.
    The output of g should be a function object of the form
    f(u, v, radians=False, inverse=False).
    For example, see the healpix() function in ``pj_healpix.py``.
    """

    def __init__(self, ellipsoid=WGS84_ELLIPSOID, proj=None, **kwargs):
        self.proj = proj
        # Keyword arguments related to the projection but not to its
        # underlying ellipsoid, e.g. for rhealpix these could be
        # {'north_square':1, 'south_square': 2}:
        self.kwargs = kwargs
        self.ellipsoid = ellipsoid

    def __str__(self):
        result = ["map projection:"]
        result.append("    proj = %s" % self.proj)
        result.append("    kwargs = %s" % self.kwargs)
        result.append("    ellipsoid:")
        for k, v in sorted(self.ellipsoid.__dict__.items()):
            result.append(" " * 8 + k + " = " + str(v))
        return "\n".join(result)

    def __call__(self, u, v, inverse=False):
        ellipsoid = self.ellipsoid
        proj = self.proj
        kwargs = self.kwargs
        lon_0 = ellipsoid.lon_0
        lat_0 = ellipsoid.lat_0
        radians = ellipsoid.radians
        a = ellipsoid.a
        e = ellipsoid.e
        if proj in HOMEMADE_PROJECTIONS:
            try:
                # Import projection module for proj.
                module = importlib.import_module("vgrid.utils.rhealpixdggs.pj_" + proj)
                f = getattr(module, proj)(a=a, e=e, **kwargs)
            except NameError:
                print("Oops! Projection %s is not implemented." % proj)
                return
        else:
            # Use a projection from the PROJ library.
            f = pyproj.Proj(proj=proj, a=a, e=e, **kwargs)
        if not inverse:
            # Translate longitudes and latitudes so that
            # (lon_0, lat_0) maps to (0, 0) in the plane.
            lam = wrap_longitude(u - lon_0, radians=radians)
            phi = wrap_latitude(v - lat_0, radians=radians)
            return f(lam, phi, radians=radians)
        else:
            lam, phi = f(u, v, radians=radians, inverse=True)
            # Translate longitudes and latitudes so that
            # (0, 0) in the plane maps to (lon_0, lat_0) on the ellipsoid.
            lam = wrap_longitude(lam + lon_0, radians=radians)
            phi = wrap_latitude(phi + lat_0, radians=radians)
            return lam, phi



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint
from vgrid.utils.eaggr.shapes.lat_long_linestring import LatLongLinestring


## Unit tests for the LatLongLinestring class.
class TestLatLongLinestring(unittest.TestCase):
    def test_getters(self):
        orig_points = [LatLongPoint(1.0, 2.0, 3.0), LatLongPoint(4.0, 5.0, 6.0)]
        linestring = LatLongLinestring()
        linestring.add_point(orig_points[0])
        linestring.add_point(orig_points[1])
        points = linestring.get_points()
        self.assertEqual(len(points), 2)
        self.assertEqual(orig_points[0], points[0])
        self.assertEqual(orig_points[1], points[1])

    def test_errors(self):
        linestring = LatLongLinestring()
        with self.assertRaises(ValueError):
            linestring.add_point("Not a LatLongPoint")



from vgrid.utils import s2, olc, geohash, georef, mgrs, maidenhead, tilecode, qtm
import h3

from vgrid.utils.gars.garsgrid import GARSGrid

from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.generator.isea3hgrid import isea3h_res_accuracy_dict
    from vgrid.generator.isea4tgrid import isea4t_res_accuracy_dict

if platform.system() == "Linux":
    from vgrid.utils.dggrid4py import DGGRIDv7, dggs_types
    import geopandas as gpd
    from vgrid.utils.dggrid4py.dggrid_runner import output_address_types

from shapely import Point

from vgrid.utils.easedggs.dggs.grid_addressing import geos_to_grid_ids

import argparse


def latlon2h3(lat, lon, res=13):
    # res: [0..15]
    if res < 0 or res > 15:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..15]."
        )
    h3_id = h3.latlng_to_cell(lat, lon, res)
    return h3_id


def latlon2h3_cli():
    """
    Command-line interface for latlon2h3.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to H3 code at a specific Resolution [0.15]. \
                                     Usage: latlon2h3 <lat> <lon> <res> [0..15]. \
                                     Ex: latlon2h3 10.775275567242561 106.70679737574993 13"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..15]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 15:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..15]."
        )
        return

    h3_id = latlon2h3(args.lat, args.lon, args.res)
    print(h3_id)


def latlon2s2(lat, lon, res=21):
    # res: [0..30]
    if res < 0 or res > 30:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..30]."
        )
    lat_lng = s2.LatLng.from_degrees(lat, lon)
    cell_id = s2.CellId.from_lat_lng(lat_lng)  # return S2 cell at max level 30
    cell_id = cell_id.parent(res)  # get S2 cell at resolution
    cell_token = s2.CellId.to_token(
        cell_id
    )  # get Cell ID Token, shorter than cell_id.id()
    return cell_token


def latlon2s2_cli():
    """
    Command-line interface for latlon2s2.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to S2 code at a specific Resolution [0..30]. \
                                     Usage: latlon2s2 <lat> <lon> <res> [0..30]. \
                                     Ex: latlon2s2 10.775275567242561 106.70679737574993 21"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..30]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 30:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..30]."
        )
        return

    s2_cell = latlon2s2(args.lat, args.lon, res)
    print(s2_cell)


def latlon2rhealpix(lat, lon, res=14):
    # res: [0..15]
    if res < 0 or res > 15:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..15]."
        )
    E = WGS84_ELLIPSOID
    rhealpix_dggs = RHEALPixDGGS(ellipsoid=E, north_square=1, south_square=3, N_side=3)
    point = (lon, lat)
    rhealpix_cell = rhealpix_dggs.cell_from_point(res, point, plane=False)
    return str(rhealpix_cell)


def latlon2rhealpix_cli():
    """
    Command-line interface for latlon2rhealpix.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to Rhealpix code at a specific Resolution [0..15]. \
                                     Usage: latlon2rhealpix <lat> <lon> <res> [0..15]. \
                                     Ex: latlon2rhealpix 10.775275567242561 106.70679737574993 14"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..15]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 15:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..15]."
        )
        return

    rhealpix_cell = latlon2rhealpix(args.lat, args.lon, res)
    print(rhealpix_cell)


def latlon2isea4t(lat, lon, res=21):
    if platform.system() == "Windows":
        # res: [0..39]
        if res < 0 or res > 39:
            raise ValueError(
                f"Invalid resolution {res}. Please input a valid resolution in [0..39]."
            )
        isea4t_dggs = Eaggr(Model.ISEA4T)
        max_accuracy = isea4t_res_accuracy_dict[
            39
        ]  # maximum cell_id length with 41 characters
        lat_long_point = LatLongPoint(lat, lon, max_accuracy)
        isea4t_cell_max_accuracy = isea4t_dggs.convert_point_to_dggs_cell(
            lat_long_point
        )
        cell_id_len = res + 2
        isea4t_cell = DggsCell(isea4t_cell_max_accuracy._cell_id[:cell_id_len])
        return isea4t_cell._cell_id


def latlon2isea4t_cli():
    """
    Command-line interface for latlon2isea4t.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to OpenEaggr ISEA4T code at a specific Resolution [0..39]. \
                                     Usage: latlon2isea4t <lat> <lon> <res> [0..39]. \
                                     Ex: latlon2isea4t 10.775275567242561 106.70679737574993 21"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..39]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 39:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..39]."
        )
        return

    eaggr_cell = latlon2isea4t(args.lat, args.lon, res)
    print(eaggr_cell)


def latlon2isea3h(lat, lon, res=27):
    if platform.system() == "Windows":
        # res: [0..40], res=27 is suitable for geocoding
        if res < 0 or res > 40:
            raise ValueError(
                f"Invalid resolution {res}. Please input a valid resolution in [0..40]."
            )
        isea3h_dggs = Eaggr(Model.ISEA3H)
        accuracy = isea3h_res_accuracy_dict.get(res)
        lat_long_point = LatLongPoint(lat, lon, accuracy)
        isea3h_cell = isea3h_dggs.convert_point_to_dggs_cell(lat_long_point)
        return isea3h_cell.get_cell_id()


def latlon2isea3h_cli():
    """
    Command-line interface for latlon2isea3h.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to OpenEaggr ISEA3H code at a specific Resolution [0..40]. \
                                     Usage: latlon2isea3h <lat> <lon> <res> [0..40]. \
                                     Ex: latlon2isea3h 10.775275567242561 106.70679737574993 14"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..40]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 40:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..40]."
        )
        return

    isea3h_cell = latlon2isea3h(args.lat, args.lon, res)
    print(isea3h_cell)


def latlon2dggrid(lat, lon, dggs_type, res, address_type="SEQNUM"):
    if platform.system() == "Linux":
        dggrid_instance = DGGRIDv7(
            executable="/usr/local/bin/dggrid",
            working_dir=".",
            capture_logs=False,
            silent=True,
            tmp_geo_out_legacy=False,
            debug=False,
        )
        point = Point(lon, lat)
        geodf_points_wgs84 = gpd.GeoDataFrame([{"geometry": point}], crs="EPSG:4326")
        dggrid_cell = dggrid_instance.cells_for_geo_points(
            geodf_points_wgs84=geodf_points_wgs84,
            cell_ids_only=True,
            dggs_type=dggs_type,
            resolution=res,
        )
        seqnum = dggrid_cell.loc[0, "seqnums"]
        address_type_transform = dggrid_instance.address_transform(
            [seqnum],
            dggs_type=dggs_type,
            resolution=res,
            mixed_aperture_level=None,
            input_address_type="SEQNUM",
            output_address_type=address_type,
        )
        # cell_id_api = {f'{dggs_type}_{res}_{address_type_transform.columns[1]}': address_type_transform.loc[0,address_type]}
        dggrid_cell_id = address_type_transform.loc[0, address_type]
        return dggrid_cell_id
        # return address_type_transform


def latlon2dggrid_cli():
    """
    Command-line interface for latlon2dggrid.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to DGGRID cell at a specific Resolution. \
                                     Usage: latlon2dggrid <lat> <lon> <dggs_type> <res>. \
                                     Ex: latlon2dggrid  10.775275567242561 106.70679737574993 ISEA7H 13"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument(
        "dggs_type",
        choices=dggs_types,
        help="Select a DGGS type from the available options.",
    )
    parser.add_argument("res", type=int, help="Resolution")
    parser.add_argument(
        "address_type",
        choices=output_address_types,
        default="SEQNUM",
        nargs="?",  # This makes the argument optional
        help="Select an output address type from the available options.",
    )

    args = parser.parse_args()
    dggs_type = args.dggs_type
    res = args.res
    address_type = args.address_type
    dggrid_cell_id = latlon2dggrid(args.lat, args.lon, dggs_type, res, address_type)
    print(dggrid_cell_id)


def latlon2ease(lat, lon, res=6):
    # res = [0..6]
    if res < 0 or res > 6:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..6]."
        )
    easedggs_cell = geos_to_grid_ids([(lon, lat)], level=res)
    easedggs_cell_id = easedggs_cell["result"]["data"][0]
    return easedggs_cell_id


def latlon2ease_cli():
    """
    Command-line interface for latlon2isea3h.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to EASE-DGGS cell at a specific Resolution [0..6]. \
                                     Usage: latlon2ease <lat> <lon> <res> [0..6]. \
                                     Ex: latlon2ease 10.775275567242561 106.70679737574993 6"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..6]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 6:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..6]."
        )
        return

    easedggs_cell = latlon2ease(args.lat, args.lon, res)
    print(easedggs_cell)


def latlon2qtm(lat, lon, res=10):
    # res: [1..24]
    if res < 1 or res > 24:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [1..24]."
        )
    return qtm.latlon_to_qtm_id(lat, lon, res)


def latlon2qtm_cli():
    """
    Command-line interface for latlon2qtm.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to QTM. \
                                     Usage: latlon2qtm <lat> <lon> <res> [1..24]. \
                                     Ex: latlon2qtm 10.775275567242561 106.70679737574993 10"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [1..24]")
    args = parser.parse_args()

    res = args.res
    if res < 1 or res > 24:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [1..24]."
        )
        return

    qtm_id = latlon2qtm(args.lat, args.lon, res)
    print(qtm_id)


def latlon2olc(lat, lon, res=11):
    # res: [2,4,6,8,10..15]
    valid_resolutions = [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]
    if res not in valid_resolutions:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in {valid_resolutions}."
        )
    olc_cell = olc.encode(lat, lon, res)
    return olc_cell


def latlon2olc_cli():
    """
    Command-line interface for latlon2olc.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to OLC/ Google Plus Code at a specific Code length [10..15]. \
                                     Usage: latlon2olc <lat> <lon> <res> [2,4,6,8,10..15]. \
                                     Ex: latlon2olc 10.775275567242561 106.70679737574993 11"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument(
        "res",
        type=int,
        choices=[2, 4, 6, 8, 10, 11, 12, 13, 14, 15],
        help="Resolution of the OLC DGGS (choose from 2, 4, 6, 8, 10, 11, 12, 13, 14, 15)",
    )

    args = parser.parse_args()

    res = args.res

    olc_cell = latlon2olc(args.lat, args.lon, res)
    print(olc_cell)


def latlon2geohash(lat, lon, res=6):
    # res: [1..10]
    if res < 1 or res > 10:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [1..10]."
        )
    geohash_id = geohash.encode(lat, lon, res)
    return geohash_id


def latlon2geohash_cli():
    """
    Command-line interface for latlon2geohash.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to Geohash code at a specific resolution [1..10]. \
                                     Usage: latlon2geohash <lat> <lon> <res>[1..10]. \
                                     Ex: latlon2geohash 10.775275567242561 106.70679737574993 6"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [1..10]")
    args = parser.parse_args()

    res = args.res
    if res < 1 or res > 10:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [1..10]."
        )
        return

    geohash_id = latlon2geohash(args.lat, args.lon, res)
    print(geohash_id)


def latlon2georef(lat, lon, res=4):
    # res: [0..5]
    if res < 0 or res > 5:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..5]."
        )
    georef_cell = georef.encode(lat, lon, res)
    return georef_cell


def latlon2georef_cli():
    """
    Command-line interface for latlon2georef.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to GEOREF code at a specific resolution [0..5]. \
                                     Usage: latlon2georef <lat> <lon> <res> [0..5]. \
                                     Ex: latlon2georef 10.775275567242561 106.70679737574993 5"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [0..5]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 5:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..5]."
        )
        return

    georef_cell = latlon2georef(args.lat, args.lon, res)
    print(georef_cell)


def latlon2mgrs(lat, lon, res=4):
    # res: [0..5]
    if res < 0 or res > 5:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..5]."
        )
    mgrs_cell = mgrs.toMgrs(lat, lon, res)
    return mgrs_cell


def latlon2mgrs_cli():
    """
    Command-line interface for latlon2mgrs.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to GEOREF code at a specific resolution [0..5]. \
                                     Usage: latlon2mgrs <lat> <lon> <res> [0..5]. \
                                     Ex: latlon2mgrs 10.775275567242561 106.70679737574993 4"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution  [0..5]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 5:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..5]."
        )
        return

    mgrs_cell = latlon2mgrs(args.lat, args.lon, res)
    print(mgrs_cell)


def latlon2tilecode(lat, lon, res=23):
    # res: [0..29]
    if res < 0 or res > 29:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..29]."
        )
    tilecode_id = tilecode.latlon2tilecode(lat, lon, res)
    return tilecode_id


def latlon2tilecode_cli():
    """
    Command-line interface for latlon2tilecode.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to Tile code at a specific resolution/ zoom level [0..29]. \
                                     Usage: latlon2tilecode <lat> <lon> <res> [0..29]. \
                                     Ex: latlon2tilecode 10.775275567242561 106.70679737574993 23"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution/ Zoom level [0..29]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 29:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..29]."
        )
        return

    tilecode_cell = latlon2tilecode(args.lat, args.lon, res)
    print(tilecode_cell)


def latlon2quadkey(lat, lon, res=23):
    # res: [0..29]
    if res < 0 or res > 29:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [0..29]."
        )
    quadkey = tilecode.latlon2quadkey(lat, lon, res)
    return quadkey


def latlon2quadkey_cli():
    """
    Command-line interface for latlon2tilecode.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to Quadkey at a specific resolution/ zoom level [0..29]. \
                                     Usage: latlon2quadkey <lat> <lon> <res> [0..29]. \
                                     Ex: latlon2quadkey 10.775275567242561 106.70679737574993 23"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution/ Zoom level [0..29]")
    args = parser.parse_args()

    res = args.res
    if res < 0 or res > 29:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolutions in [0..29]."
        )
        return

    quadkey = latlon2quadkey(args.lat, args.lon, res)
    print(quadkey)


def latlon2maidenhead(lat, lon, res=4):
    # res: [1..4]
    if res < 1 or res > 4:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [1..4]."
        )
    maidenhead_cell = maidenhead.toMaiden(lat, lon, res)
    return maidenhead_cell


def latlon2maidenhead_cli():
    """
    Command-line interface for latlon2maidenhead.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to Tile code at a specific resolution [1..4]. \
                                     Usage: latlon2maidenhead <lat> <lon> <res> [1..4]. \
                                     Ex: latlon2maidenhead 10.775275567242561 106.70679737574993 4"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument("res", type=int, help="Input Resolution [1..4]")
    args = parser.parse_args()

    res = args.res
    if res < 1 or res > 4:
        print(
            f"Error: Invalid resolution {args.res}. Please input a valid resolutions in [1..4]."
        )
        return

    maidenhead_cell = latlon2maidenhead(args.lat, args.lon, res)
    print(maidenhead_cell)


def latlon2gars(lat, lon, res=1):
    # res: [1..4] where 1 is min res (30 minutes), 4 is max res (1 minute)
    if res < 1 or res > 4:
        raise ValueError(
            f"Invalid resolution {res}. Please input a valid resolution in [1..4]."
        )
    # Convert res to minutes: 1->30, 2->15, 3->5, 4->1
    minutes_map = {1: 30, 2: 15, 3: 5, 4: 1}
    minutes = minutes_map[res]
    gars_cell = GARSGrid.from_latlon(lat, lon, minutes)
    return str(gars_cell)


def latlon2gars_cli():
    """
    Command-line interface for latlon2gars.
    """
    parser = argparse.ArgumentParser(
        description="Convert Lat, Long to GARS code at a specific resolution [1..4]. \
                                     Usage: latlon2gars <lat> <lon> <res> [1..4]. \
                                     Ex: latlon2gars 10.775275567242561 106.70679737574993 1"
    )
    parser.add_argument("lat", type=float, help="Input Latitude")
    parser.add_argument("lon", type=float, help="Input Longitude")
    parser.add_argument(
        "res",
        type=int,
        help="Input Resolution [1..4] (1=30min, 2=15min, 3=5min, 4=1min)",
    )
    args = parser.parse_args()

    res = args.res
    if res < 1 or res > 4:
        print(
            f"Error: Invalid resolution {res}. Please input a valid resolution in [1..4]."
        )
        return

    gars_cell = latlon2gars(args.lat, args.lon, res)
    print(gars_cell)



import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.utils import qtm
from vgrid.generator.settings import geodesic_dggs_to_feature
from vgrid.conversion.dggscompact import qtmcompact
from vgrid.utils.geometry import (
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)
from vgrid.stats.qtmstats import qtm_metrics

# QTM facet points
p90_n180, p90_n90, p90_p0, p90_p90, p90_p180 = (
    (90.0, -180.0),
    (90.0, -90.0),
    (90.0, 0.0),
    (90.0, 90.0),
    (90.0, 180.0),
)
p0_n180, p0_n90, p0_p0, p0_p90, p0_p180 = (
    (0.0, -180.0),
    (0.0, -90.0),
    (0.0, 0.0),
    (0.0, 90.0),
    (0.0, 180.0),
)
n90_n180, n90_n90, n90_p0, n90_p90, n90_p180 = (
    (-90.0, -180.0),
    (-90.0, -90.0),
    (-90.0, 0.0),
    (-90.0, 90.0),
    (-90.0, 180.0),
)


def validate_qtm_resolution(resolution):
    """
    Validate that QTM resolution is in the valid range [1..24].

    Args:
        resolution (int): Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [1..24]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 1 or resolution > 24:
        raise ValueError(f"Resolution must be in range [1..24], got {resolution}")

    return resolution


def point2qtm(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,  # New parameter for topology preservation
):
    """
    Convert a point geometry to a QTM grid cell.

    Args:
        resolution (int): QTM resolution [1..24]
        point (shapely.geometry.Point): Point geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for points)
        compact (bool, optional): Enable QTM compact mode (not used for points)
        topology (bool, optional): Enable topology preserving mode - ensures disjoint points have disjoint QTM cells
        include_properties (bool, optional): Whether to include properties in output
        all_points: List of all points for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing QTM cells containing the point
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        
        # Calculate the shortest distance between all points
        shortest_distance = shortest_point_distance(all_points)
        
        # Find resolution where QTM cell size is smaller than shortest distance
        # This ensures disjoint points have disjoint QTM cells
        if shortest_distance > 0:
            for res in range(1, 25):  # QTM resolution range is [1..24]
                _, avg_edge_length, _ = qtm_metrics(res)
                # Use a factor to ensure sufficient separation (triangle diameter is ~2x edge length)
                triangle_diameter = avg_edge_length * 2
                if triangle_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 24
        else:
            # Single point or no distance, use provided resolution
            pass

    qtm_features = []
    latitude = point.y
    longitude = point.x
    qtm_id = qtm.latlon_to_qtm_id(latitude, longitude, resolution)
    facet = qtm.qtm_id_to_facet(qtm_id)
    cell_polygon = qtm.constructGeometry(facet)
    if cell_polygon:
        num_edges = 3
        qtm_feature = geodesic_dggs_to_feature(
            "qtm", qtm_id, resolution, cell_polygon, num_edges
        )
        if include_properties and feature_properties:
            qtm_feature["properties"].update(feature_properties)
        qtm_features.append(qtm_feature)
    return qtm_features


def polyline2qtm(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,  # New parameter for topology preservation
):
    """
    Convert line geometries (LineString, MultiLineString) to QTM grid cells.

    Args:
        resolution (int): QTM resolution [1..24]
        feature (shapely.geometry.LineString or shapely.geometry.MultiLineString): Line geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for lines)
        compact (bool, optional): Enable QTM compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polylines have disjoint QTM cells
        include_properties (bool, optional): Whether to include properties in output
        all_polylines: List of all polylines for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing QTM cells intersecting the line
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        
        # Calculate the shortest distance between all polylines
        shortest_distance = shortest_polyline_distance(all_polylines)
        
        # Find resolution where QTM cell size is smaller than shortest distance
        # This ensures disjoint polylines have disjoint QTM cells
        if shortest_distance > 0:
            for res in range(1, 25):  # QTM resolution range is [1..24]
                _, avg_edge_length, _ = qtm_metrics(res)
                # Use a factor to ensure sufficient separation (triangle diameter is ~2x edge length)
                triangle_diameter = avg_edge_length * 4  # in case there are 2 cells representing the same line segment
                if triangle_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 24
        else:
            # Single polyline or no distance, use provided resolution
            pass
    qtm_features = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)
    else:
        return []
    for polyline in polylines:
        levelFacets = {}
        QTMID = {}
        for lvl in range(resolution):
            levelFacets[lvl] = []
            QTMID[lvl] = []
            if lvl == 0:
                initial_facets = [
                    [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
                    [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
                    [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
                    [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
                    [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
                    [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
                    [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
                    [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
                ]
                for i, facet in enumerate(initial_facets):
                    QTMID[0].append(str(i + 1))
                    levelFacets[0].append(facet)
                    facet_geom = qtm.constructGeometry(facet)
                    if Polygon(facet_geom).intersects(polyline) and resolution == 1:
                        qtm_id = QTMID[0][i]
                        num_edges = 3
                        qtm_feature = geodesic_dggs_to_feature(
                            "qtm", qtm_id, resolution, facet_geom, num_edges
                        )
                        if include_properties and feature_properties:
                            qtm_feature["properties"].update(feature_properties)
                        qtm_features.append(qtm_feature)
                        return qtm_features
            else:
                for i, pf in enumerate(levelFacets[lvl - 1]):
                    subdivided_facets = qtm.divideFacet(pf)
                    for j, subfacet in enumerate(subdivided_facets):
                        subfacet_geom = qtm.constructGeometry(subfacet)
                        if Polygon(subfacet_geom).intersects(polyline):
                            new_id = QTMID[lvl - 1][i] + str(j)
                            QTMID[lvl].append(new_id)
                            levelFacets[lvl].append(subfacet)
                            if lvl == resolution - 1:
                                num_edges = 3
                                qtm_feature = geodesic_dggs_to_feature(
                                    "qtm", new_id, resolution, subfacet_geom, num_edges
                                )
                                if include_properties and feature_properties:
                                    qtm_feature["properties"].update(feature_properties)
                                qtm_features.append(qtm_feature)
    if compact:
        qtm_geojson = {"type": "FeatureCollection", "features": qtm_features}
        return qtmcompact(qtm_geojson)["features"]

    return qtm_features


def polygon2qtm(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,  # New parameter for topology preservation
):
    """
    Convert polygon geometries (Polygon, MultiPolygon) to QTM grid cells.

    Args:
        resolution (int): QTM resolution [1..24]
        feature (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): Polygon geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply ('intersect', 'within', 'centroid_within', 'largest_overlap')
        compact (bool, optional): Enable QTM compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polygons have disjoint QTM cells
        include_properties (bool, optional): Whether to include properties in output
        all_polygons: List of all polygons for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing QTM cells based on predicate
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        
        # Calculate the shortest distance between all polygons
        shortest_distance = shortest_polygon_distance(all_polygons)
        print(shortest_distance)
        # Find resolution where QTM cell size is smaller than shortest distance
        # This ensures disjoint polygons have disjoint QTM cells
        if shortest_distance > 0:
            for res in range(1, 25):  # QTM resolution range is [1..24]
                _, avg_edge_length, _ = qtm_metrics(res)
                # Use a factor to ensure sufficient separation (triangle diameter is ~2x edge length)
                triangle_diameter = avg_edge_length * 4
                if triangle_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 24
        else:
            # Single polygon or no distance, use provided resolution
            pass
    qtm_features = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)
    else:
        return []
    for polygon in polygons:
        levelFacets = {}
        QTMID = {}
        for lvl in range(resolution):
            levelFacets[lvl] = []
            QTMID[lvl] = []
            if lvl == 0:
                initial_facets = [
                    [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
                    [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
                    [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
                    [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
                    [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
                    [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
                    [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
                    [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
                ]
                for i, facet in enumerate(initial_facets):
                    QTMID[0].append(str(i + 1))
                    levelFacets[0].append(facet)
                    facet_geom = qtm.constructGeometry(facet)
                    if Polygon(facet_geom).intersects(polygon) and resolution == 1:
                        qtm_id = QTMID[0][i]
                        num_edges = 3
                        qtm_feature = geodesic_dggs_to_feature(
                            "qtm", qtm_id, resolution, facet_geom, num_edges
                        )
                        if include_properties and feature_properties:
                            qtm_feature["properties"].update(feature_properties)
                        qtm_features.append(qtm_feature)
                        return qtm_features
            else:
                for i, pf in enumerate(levelFacets[lvl - 1]):
                    subdivided_facets = qtm.divideFacet(pf)
                    for j, subfacet in enumerate(subdivided_facets):
                        subfacet_geom = qtm.constructGeometry(subfacet)
                        if Polygon(subfacet_geom).intersects(polygon):
                            new_id = QTMID[lvl - 1][i] + str(j)
                            QTMID[lvl].append(new_id)
                            levelFacets[lvl].append(subfacet)
                            if lvl == resolution - 1:
                                if not check_predicate(
                                    Polygon(subfacet_geom), polygon, predicate
                                ):
                                    continue
                                num_edges = 3
                                qtm_feature = geodesic_dggs_to_feature(
                                    "qtm", new_id, resolution, subfacet_geom, num_edges
                                )
                                if include_properties and feature_properties:
                                    qtm_feature["properties"].update(feature_properties)
                                qtm_features.append(qtm_feature)
    if compact:
        qtm_geojson = {"type": "FeatureCollection", "features": qtm_features}
        return qtmcompact(qtm_geojson)["features"]
    return qtm_features


def geometry2qtm(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a list of geometries to QTM grid cells.

    Args:
        geometries (shapely.geometry.BaseGeometry or list): Single geometry or list of geometries
        resolution (int): QTM resolution [1..24]
        properties_list (list, optional): List of property dictionaries for each geometry
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable QTM compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with QTM grid cells
    """
    resolution = validate_qtm_resolution(resolution)

    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    qtm_features = []
    for idx, geom in enumerate(tqdm(geometries, desc="Processing features")):
        props = (
            properties_list[idx]
            if properties_list and idx < len(properties_list)
            else {}
        )
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            qtm_features.extend(
                point2qtm(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                qtm_features.extend(
                    point2qtm(
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,  # Pass all points for topology preservation
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            qtm_features.extend(
                polyline2qtm(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,  # Pass all polylines for topology preservation
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            qtm_features.extend(
                polygon2qtm(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons=all_polygons,  # Pass all polygons for topology preservation
                )
            )
    return {"type": "FeatureCollection", "features": qtm_features}


def dataframe2qtm(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a pandas DataFrame with geometry column to QTM grid cells.

    Args:
        df (pandas.DataFrame): DataFrame with geometry column
        resolution (int): QTM resolution [1..24]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable QTM compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with QTM grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2qtm(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2qtm(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a GeoDataFrame to QTM grid cells.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to convert
        resolution (int): QTM resolution [1..24]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable QTM compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with QTM grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2qtm(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def vector2qtm(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to QTM grid cells from various input formats.

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): QTM resolution [1..24]
        compact (bool): Enable QTM compact mode for polygons (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions
    Returns:
        dict or str: Output in the specified format
    """
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        result = geodataframe2qtm(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        result = dataframe2qtm(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        result = geometry2qtm(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2qtm(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2qtm(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert QTM result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since QTM uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2qtm.gpkg", driver="GPKG")
            return "vector2qtm.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2qtm.parquet", index=False)
            return "vector2qtm.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2qtm.shp", driver="ESRI Shapefile")
            return "vector2qtm.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2qtm_cli():
    """
    Command-line interface for vector2qtm conversion.

    Usage:
        python vector2qtm.py -i input.shp -r 10 -c -f geojson -o output.geojson

    Arguments:
        -i, --input: Input file path or URL
        -r, --resolution: QTM resolution [1..24]
        -c, --compact: Enable QTM compact mode
        -p, --predicate: Spatial predicate (intersect, within, centroid_within, largest_overlap)
        -t, --topology: Enable topology preserving mode
        -np, --no-props: Do not include original feature properties
        -f, --format: Output format (geojson, gpkg, parquet, csv, shapefile)
        -o, --output: Output file path
    """
    parser = argparse.ArgumentParser(
        description="Convert vector data to QTM grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(1, 25),
        metavar="[1-24]",
        help="QTM resolution [1..24] (1=coarsest, 24=finest)",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable QTM compact mode for polygons",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_qtm_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2qtm{extensions.get(args.format, '')}"
    try:
        vector2qtm(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2qtm_cli()



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 15 15:00:27 2018

@author: Joseph
"""

"""
This Python 3.3 module implements the quadrilateralized spherical cube (QSC) map projection as described in [CaGr2002]_.
.. [CaGr2002] M. R. Calabretta and E. W. Greisen, Representations of celestial coordinates in FITS, Astronomy & Astrophysics 395 (2002), 1077--1122.
CHANGELOG:
- Alexander Raichev (AR), 2012-02-04: Initial version.
- AR, 2013-07-23: Ported to Python 3.3.
NOTE:
All lengths are measured in meters and all angles are measured in radians 
unless indicated otherwise. 
By 'ellipsoid' below, i mean an oblate ellipsoid of revolution.
"""
# *****************************************************************************
#       Copyright (C) 2012 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import third-party modules.
from numpy import pi, rad2deg, deg2rad, sin, cos, argmax, sqrt, arctan, arcsin, array

# Import my modules.
from .utils import auth_lat, auth_rad

face_centers_lonlat = [
    (0, pi / 2),
    (0, 0),
    (pi / 2, 0),
    (pi, 0),
    (3 * pi / 2, 0),
    (0, -pi / 2),
]


def qsc_sphere(lam, phi):
    r"""
    Compute the signature function of the icosohedral Snyder equal area
    (ISEA) map projection of the unit sphere.

    INPUT:

    - `lam, phi` - Geodetic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.
    """
    # Find out what face (lam, phi) lies on.
    L = cos(lam) * cos(phi)
    M = sin(lam) * cos(phi)
    N = sin(phi)
    face = argmax([N, L, M, -L, -M, -N])
    if face == 0:
        xi, eta, zeta = M, -L, N
    elif face == 1:
        xi, eta, zeta = M, N, L
    elif face == 2:
        xi, eta, zeta = -L, N, M
    elif face == 3:
        xi, eta, zeta = -M, N, -L
    elif face == 4:
        xi, eta, zeta = L, N, -M
    else:
        # face = 5
        xi, eta, zeta = M, L, -N
    lam_0, phi_0 = face_centers_lonlat[face]
    if xi > abs(eta) or eta > abs(xi):
        S = 1
    else:
        S = -1
    if abs(xi) > abs(eta):
        omega = eta / xi
    else:
        if eta == 0:
            omega = 0
        else:
            omega = xi / eta
    u = (pi / 4) * S * sqrt((1 - zeta) / (1 - 1 / sqrt(2 + omega**2)))
    v = u / (pi / 12) * (arctan(omega) - arcsin(omega / (sqrt(2 * (1 + omega**2)))))
    if abs(xi) > abs(eta):
        x, y = array([lam_0, phi_0]) + array([u, v])
    else:
        x, y = array([lam_0, phi_0]) + array([v, u])
    return x, y


def qsc_sphere_inverse(x, y):
    r"""
    Compute the inverse of qsc_sphere().
    INPUT:

    - `x, y` - Planar coordinates in meters in the image of the
      ISEA projection of the unit sphere.

    NOTES:

    Not implemented here nor in PROJ.4.
    """
    pass


def qsc_ellipsoid(lam, phi, e=0):
    r"""
    Compute the signature functions of the ISEA may projection of an oblate
    ellipsoid with eccentricity `e` whose authalic sphere is the unit sphere.

    INPUT:

    - `lam, phi` - Geodetic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.
    - `e` - Eccentricity of the oblate ellipsoid.
    """
    beta = auth_lat(phi, e, radians=True)
    return qsc_sphere(lam, beta)


def qsc_ellipsoid_inverse(x, y, e=0):
    r"""
    Compute the inverse of qsc_ellipsoid().

    INPUT:

    - `x, y` - Planar coordinates in meters in the image of the
      ISEA projection of the ellipsoid with eccentricity `e` whose
      authalic sphere is the unit sphere.

    NOTES:

    Not implemented here nor in PROJ.4.
    """
    pass


def qsc(a=1, e=0):
    r"""
    Return a function object that wraps the ISEA projection and its inverse
    of an ellipsoid with major radius `a` and eccentricity `e`.

    OUTPUT:

    - A function object of the form f(u, v, radians=False, inverse=False).
    """
    R_A = auth_rad(a, e)

    def f(u, v, radians=False, inverse=False):
        if not inverse:
            lam, phi = u, v
            if not radians:
                # Convert to radians.
                lam, phi = deg2rad([lam, phi])
            return tuple(R_A * array(qsc_ellipsoid(lam, phi, e=e)))
        else:
            # Scale down to R_A = 1.
            x, y = array((u, v)) / R_A
            lam, phi = array(qsc_ellipsoid_inverse(x, y, e=e))
            if not radians:
                # Convert to degrees.
                lam, phi = rad2deg([lam, phi])
            return lam, phi

    return f



import json
import os
import argparse
from vgrid.utils.antimeridian import fix_geojson


def fix_geom(input_file, output_file):
    with open(input_file, "r") as infile:
        geojson_data = json.load(infile)  # Load as dictionary

    fixed_geojson_data = fix_geojson(geojson_data)
    with open(output_file, "w") as outfile:
        json.dump(fixed_geojson_data, outfile, indent=2)

    print(f"GeoJSON has been fixed antimeridian and saved to {output_file}")


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(
        description="Fix antimeridian GeoJSON created from DGGRID"
    )
    parser.add_argument("input", help="Input GeoJSON file")

    # Parse the arguments
    args = parser.parse_args()

    # Derive output filename by appending "_fixed" to the input filename
    # output_file = f"{args.input.rsplit('.', 1)[0]}_fixed.geojson"
    output_file = os.path.join(
        os.getcwd(),
        f"{os.path.splitext(os.path.basename(args.input))[0]}_geom_fixed.geojson",
    )

    # Call the function with the input and output file
    fix_geom(args.input, output_file)


if __name__ == "__main__":
    main()



import locale
import argparse
import csv
import math
from texttable import Texttable

locale.setlocale(locale.LC_ALL, "")


def tilecode_metrics(res):
    earth_surface_area_km2 = 510_065_621.724  # 510.1 million square kilometers
    num_cells = 4**res
    avg_area = (earth_surface_area_km2 / num_cells) * (10**6)
    avg_edge_length = math.sqrt(avg_area)
    return num_cells, avg_edge_length, avg_area


def tilecode_stats(output_file=None):
    min_res = 0
    max_res = 29

    # Create a Texttable object for displaying in the terminal
    t = Texttable()

    # Add header to the table, including the new 'Cell Width' and 'Cell Area' columns
    t.add_row(
        ["Resolution", "Number of Cells", "Avg Edge Length (m)", "Avg Cell Area (sq m)"]
    )

    # Check if an output file is specified (for CSV export)
    if output_file:
        # Open the output CSV file for writing
        with open(output_file, mode="w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(
                [
                    "Resolution",
                    "Number of Cells",
                    "Avg Edge Length (m)",
                    "Avg Cell Area (sq m)",
                ]
            )
            for res in range(min_res, max_res + 1):
                num_cells, avg_edge_length, avg_area = tilecode_metrics(res)
                writer.writerow([res, num_cells, avg_edge_length, avg_area])
    else:
        for res in range(min_res, max_res + 1):
            num_cells, avg_edge_length, avg_area = tilecode_metrics(res)
            formatted_num_cells = locale.format_string("%d", num_cells, grouping=True)
            formatted_length = locale.format_string(
                "%.3f", avg_edge_length, grouping=True
            )
            formatted_area = locale.format_string("%.3f", avg_area, grouping=True)

            t.add_row([res, formatted_num_cells, formatted_length, formatted_area])

        print(t.draw())


def main():
    # Set up command-line argument parsing
    parser = argparse.ArgumentParser(description="Export or display Tilecode stats.")
    parser.add_argument("-o", "--output", help="Output CSV file name.")
    args = parser.parse_args()

    # Call the function with the provided output file (if any)
    tilecode_stats(args.output)


if __name__ == "__main__":
    main()
